#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/gpio.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/platform_device.h>
#include <linux/of_gpio.h>

MODULE_LICENSE("Dual MIT/GPL");
MODULE_AUTHOR("Marcin Bis <marcin@bis-linux.com>");
MODULE_DESCRIPTION("IO latency test module - in-kernel version.");

/* structure describing single inout_user device */
struct inout_user_device {
	long counter;
	int ipin; /* input GPIO pin */
	int opin; /* output GPIO pin */
};

/* interrupt handler */
static irqreturn_t inout_isr(int irq, void *dev_id)
{
	int val;
	struct inout_user_device *inout_dev = dev_id;

	inout_dev->counter++;

	val = gpio_get_value(inout_dev->ipin);
	gpio_set_value(inout_dev->opin,val);

	return IRQ_HANDLED;
}

int inout_user_probe (struct platform_device * pdev)
{
	int irq, error;
	struct inout_user_device *inout_dev;
	struct device_node *np = pdev->dev.of_node;

	/* Create new inout_dev structure */
	inout_dev = devm_kmalloc(&pdev->dev, sizeof(struct inout_user_device), GFP_KERNEL);

	inout_dev->counter = 0;

	/* Get GPIO from device-tree */
	inout_dev->ipin = of_get_named_gpio(np, "input-gpio", 0);
	inout_dev->opin = of_get_named_gpio(np, "output-gpio", 0);

	/* claim and configure output pin */
	error = gpio_request(inout_dev->opin, pdev->name);
	if (error < 0) {
		dev_err(&pdev->dev, "Failed to request GPIO %d, error %d\n",inout_dev->opin,error);
		goto err0;
	}

	error = gpio_direction_output(inout_dev->opin,1);
	if (error < 0) {
		dev_err(&pdev->dev, "Failed to set direction of GPIO %d, error %d\n",inout_dev->opin,error);
		goto err1;
	}

	gpio_set_value(inout_dev->opin,0);

	/* claim and configure input pin */
	error = gpio_request(inout_dev->ipin, pdev->name);
	if (error < 0) {
		dev_err(&pdev->dev, "Failed to request GPIO %d, error %d\n",inout_dev->ipin,error);
		goto err1;
	}

	error = gpio_direction_input(inout_dev->ipin);
	if (error < 0) {
		dev_err(&pdev->dev, "Failed to set direction of GPIO %d, error %d\n",inout_dev->ipin,error);
		goto err2;
	}

	irq = gpio_to_irq(inout_dev->ipin);
	if (error < 0) {
		error = irq;
		dev_err(&pdev->dev, "Failed to get IRQ number for GPIO %d, error %d\n",inout_dev->ipin,error);
		goto err2;
	}

	/* Store pointer to structure representing device data */
	dev_set_drvdata(&pdev->dev, inout_dev);

	/* request interrupt on input pin - the fun starts from now */
	error = request_irq(irq, inout_isr,
                            IRQF_SHARED | IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
                            pdev->name, inout_dev);

	if (error) {
		dev_err(&pdev->dev, "Failed to request interrupt %d, error %d\n",irq,error);
		goto err2;
	}

	return 0;

	err2:
		gpio_free(inout_dev->ipin);
	err1:
		gpio_free(inout_dev->opin);
	err0:
		kfree(inout_dev);
		return error;
}

int inout_user_remove (struct platform_device * pdev)
{
	int irq;
	struct inout_user_device *inout_dev;

	inout_dev = dev_get_drvdata(&pdev->dev);

	irq = gpio_to_irq(inout_dev->ipin);

	free_irq(irq, inout_dev);
	gpio_free(inout_dev->ipin);
	gpio_free(inout_dev->opin);
	dev_info(&pdev->dev, "Event was triggered %ld times\n", inout_dev->counter);
	devm_kfree(&pdev->dev, inout_dev);
	return 0;
}

static const struct of_device_id cinout_of_match[] = {
	{ .compatible = "bis,inout_user" },
	{},
};
MODULE_DEVICE_TABLE(of, cinout_of_match);

static struct platform_driver cinout_driver = {
	.probe = inout_user_probe,
	.remove = inout_user_remove,
	.driver = {
		.owner = THIS_MODULE,
		.name	= "inout_irq",
		.of_match_table = of_match_ptr(cinout_of_match),
	},
};

module_platform_driver(cinout_driver);
