#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/gpio.h>
#include <linux/workqueue.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/slab.h>
#include <linux/platform_device.h>
#include <linux/of_gpio.h>
#include <linux/miscdevice.h>

/* copy_from/to_user */
#include <linux/uaccess.h>

/* waitqueue */
#include <linux/sched.h>

MODULE_LICENSE("Dual MIT/GPL");
MODULE_AUTHOR("Marcin Bis <marcin@bis-linux.com>");
MODULE_DESCRIPTION("IO latency test module - userspace connector.");

/* waitqueue - used for userspace - interrupt synchronization */
DECLARE_WAIT_QUEUE_HEAD(inout_waitqueue);

/* structure describing inout_user device */
struct inout_user_device {
	long counter;
	int ipin; /* input GPIO pin */
	int opin; /* output GPIO pin */

	struct file_operations inout_fops;
	struct miscdevice inout_misc;
	char node_name[20];
};

/* structure used for deferred interrupt handling (BH) */
struct uinout_data {
	struct work_struct work;
	struct inout_user_device * data;
};

/* workqueue (single queue for all instances) */
static struct workqueue_struct *inout_wq;

/* BH */
static void inout_job_function(struct work_struct * work)
{
	struct uinout_data * job = 
		container_of(work, struct uinout_data, work);

	struct inout_user_device *inout_dev = job->data;

	inout_dev->counter++;

	wake_up(&inout_waitqueue);

	kfree(job);
}

/* interrupt handler */
static irqreturn_t inout_isr(int irq, void *dev_id)
{
	int ret;
	struct uinout_data * job;
	struct inout_user_device *inout_dev = dev_id;

	job = kmalloc(sizeof(struct uinout_data), GFP_ATOMIC);
	INIT_WORK(&(job->work), inout_job_function);
	job->data = inout_dev;

	ret = queue_work(inout_wq, &(job->work));

	return IRQ_HANDLED;
}

static ssize_t inout_user_read(struct file *file, char __user * userbuf, size_t count, loff_t * ppos)
{
	const struct file_operations * fops = file->f_op;
	struct inout_user_device * i_dev;
	long cur;

	i_dev = container_of(fops, struct inout_user_device, inout_fops);

	cur = i_dev->counter;
	wait_event_interruptible(inout_waitqueue, (i_dev->counter > cur));
	if (signal_pending(current))
	{
		return -EINTR;
	}
	return 0;
}

static ssize_t inout_user_write(struct file *file, const char __user * userbuf, size_t count, loff_t * ppos)
{
	const struct file_operations * fops = file->f_op;
	struct inout_user_device * i_dev;
	char tmpbuf[2] = "0";

	int val = 0;
	int res;

	i_dev = container_of(fops, struct inout_user_device, inout_fops);

	res = copy_from_user(tmpbuf, userbuf, 1);
	val = simple_strtol(tmpbuf, NULL, 10);

	//printk(KERN_INFO "Requested value: %d\n",val);

	gpio_set_value(i_dev->opin,val);
	return count;
}

int inout_user_probe (struct platform_device * pdev)
{
	int irq, error;
	struct inout_user_device *inout_dev;
	struct device_node *np = pdev->dev.of_node;
	static int counter = 0;

	/* Create new inout_dev structure
	 * use kzalloc, to get zeroed memory - spurious function pointers in other case! */
	inout_dev = devm_kzalloc(&pdev->dev, sizeof(struct inout_user_device), GFP_KERNEL);

	inout_dev->counter = 0;

	/* Get GPIO from device-tree */
	inout_dev->ipin = of_get_named_gpio(np, "input-gpio", 0);
	inout_dev->opin = of_get_named_gpio(np, "output-gpio", 0);

	/* claim and configure output pin */
	error = gpio_request(inout_dev->opin, pdev->name);
	if (error < 0) {
		dev_err(&pdev->dev, "Failed to request GPIO %d, error %d\n",inout_dev->opin,error);
		goto err0;
	}

	error = gpio_direction_output(inout_dev->opin,1);
	if (error < 0) {
		dev_err(&pdev->dev, "Failed to set direction of GPIO %d, error %d\n",inout_dev->opin,error);
		goto err1;
	}

	gpio_set_value(inout_dev->opin,0);

	/* create workqueue to handle jobs */
	inout_wq = create_workqueue("inout_wq");

	/* claim and configure input pin */
	error = gpio_request(inout_dev->ipin, pdev->name);
	if (error < 0) {
		dev_err(&pdev->dev, "Failed to request GPIO %d, error %d\n",inout_dev->ipin,error);
		goto err1;
	}

	error = gpio_direction_input(inout_dev->ipin);
	if (error < 0) {
		dev_err(&pdev->dev, "Failed to set direction of GPIO %d, error %d\n",inout_dev->ipin,error);
		goto err2;
	}

	irq = gpio_to_irq(inout_dev->ipin);
	if (error < 0) {
		error = irq;
		dev_err(&pdev->dev, "Failed to get IRQ number for GPIO %d, error %d\n",inout_dev->ipin,error);
		goto err2;
	}

	/* Store pointer to structure representing device data */
	dev_set_drvdata(&pdev->dev, inout_dev);

	/* Register misc device */
	sprintf(inout_dev->node_name, "ioexp%d", counter++);

	inout_dev->inout_fops.owner = THIS_MODULE;
	inout_dev->inout_fops.read = inout_user_read;
	inout_dev->inout_fops.write = inout_user_write;

	inout_dev->inout_misc.name = inout_dev->node_name;
	inout_dev->inout_misc.minor = MISC_DYNAMIC_MINOR;
	inout_dev->inout_misc.fops = &inout_dev->inout_fops;

	misc_register(&inout_dev->inout_misc);

	/* request interrupt on input pin - the fun starts from now */
	error = request_irq(irq, inout_isr,
                            IRQF_SHARED | IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
                            pdev->name, inout_dev);

	if (error) {
		dev_err(&pdev->dev, "Failed to request interrupt %d, error %d\n",irq,error);
		goto err2;
	}

	return 0;

	err2:
		gpio_free(inout_dev->ipin);
	err1:
		gpio_free(inout_dev->opin);
	err0:
		kfree(inout_dev);
		return error;
}

int inout_user_remove (struct platform_device * pdev)
{
	int irq;
	struct inout_user_device *inout_dev;

	inout_dev = dev_get_drvdata(&pdev->dev);

	misc_deregister(&inout_dev->inout_misc);

	irq = gpio_to_irq(inout_dev->ipin);

	free_irq(irq, inout_dev);
	flush_workqueue(inout_wq);
	destroy_workqueue(inout_wq);
	gpio_free(inout_dev->ipin);
	gpio_free(inout_dev->opin);
	dev_info(&pdev->dev, "Event was trigerred %ld times\n", inout_dev->counter);
	devm_kfree(&pdev->dev, inout_dev);
	return 0;
}

static const struct of_device_id inout_of_match[] = {
	{ .compatible = "bis,inout_user" },
	{},
};
MODULE_DEVICE_TABLE(of, inout_of_match);

static struct platform_driver inout_driver = {
	.probe = inout_user_probe,
	.remove = inout_user_remove,
	.driver = {
		.owner = THIS_MODULE,
		.name	= "inout_user",
		.of_match_table = of_match_ptr(inout_of_match),
	},
};

module_platform_driver(inout_driver);
